import struct


class Encoding:
    END         = b"\x0A"
    ESC         = b"\x0B"
    ESC_END     = b"\x1A"
    ESC_ESC     = b"\x1B"

class Serializer:
    def __init__(self):
        pass

    def _receive(self, size):
        pass

    def _transmit(self, data, size):
        pass

    def _flush(self):
        pass

    def set_receive_timeout(self, timeout):
        pass

    def set_transmit_timeout(self, timeout):
        pass

    """
    Receive a packet from serial.

    @type timeout: float
    @param timeout: 
    @return: the data received, or b"" if no data is available
    """
    def receive(self, timeout=None):
        data = b""
        c = self._receive(1)

        while c != Encoding.END:
            if c == Encoding.ESC:
                c = self._receive(1)
                if c == Encoding.ESC_END:
                    data += Encoding.END
                elif c == Encoding.ESC_ESC:
                    data += Encoding.ESC
                else:
                    data += c
            else:
                data += c
            c = self._receive(1)
        return data

    """
    Transmit a packet to serial.

    @type data: bytes
    @param data: the data that will be sent.
    """
    def transmit(self, data):
        if type(data) == str:
            raise AttributeError("cannot transmit \"str\" object. Perhaps try \"str.encode()\" it?")

        index = 0
        while index < len(data):
            c = struct.pack("B", data[index])
            if c == Encoding.END:
                self._transmit(Encoding.ESC, 1)
                self._transmit(Encoding.ESC_END, 1)
            elif c == Encoding.ESC:
                self._transmit(Encoding.ESC, 1)
                self._transmit(Encoding.ESC_ESC, 1)
            else:
                self._transmit(c, 1)
            index += 1
        self._transmit(Encoding.END, 1)

        self._flush()



import serial
import serial.tools.list_ports

from .serializer import Serializer


class UARTSerializer(Serializer):
    def __init__(self, port=None, baudrate=115200):
        self._port = port
        self.baudrate = baudrate

        if not self._port:
            # if no port is supplied, we enumerate the available ports
            # and try the first one
            ports = UARTSerializer.get_available_ports()
            if ports:
                self._port = ports[0]
                print("No \"port\" argument supplied, using {0}.".format(self._port))

        self._ser = serial.Serial(
            port=self._port, baudrate=baudrate, timeout=None)

    @staticmethod
    def get_available_ports():
        ports = [p.name for p in serial.tools.list_ports.comports()]
        return ports
        
    """
    timeout = None: wait forever / until requested number of bytes are received
    timeout = 0: non-blocking mode, return immediately in any case, returning zero or more, up to the requested number of bytes
    timeout = x: set timeout to x seconds (float allowed) returns immediately when the requested number of bytes are available, otherwise wait until the timeout expires and return all bytes that were received until then.
    """
    def set_receive_timeout(self, timeout):
        self._ser.timeout = timeout
        
    def set_transmit_timeout(self, timeout):
        self._ser.write_timeout = timeout

    def _receive(self, size):
        return self._ser.read(size)
    
    def _transmit(self, data, size):
        self._ser.write(data)




from .serializer import Serializer


class SocketSerializer(Serializer):
    def __init__(self, connection, buffered_transmit=False):
        self._conn = connection
        self.buffered_transmit = buffered_transmit
        self.tx_buffer = b""

    """
    Set a timeout on blocking socket operations. 
    
    The value argument can be a nonnegative floating point number expressing seconds, or None. 
    If a non-zero value is given, subsequent socket operations will raise a timeout exception if the timeout period value has elapsed before the operation has completed. 
    If zero is given, the socket is put in non-blocking mode. 
    If None is given, the socket is put in blocking mode.
    """
    def set_receive_timeout(self, timeout):
        self._conn.settimeout(timeout)

    def _receive(self, size, timeout=None):
        try:
            data = self._conn.recv(size)
            return data
        except TimeoutError:
            return b""
    
    def _transmit(self, data, size):
        if self.buffered_transmit:
            self.tx_buffer += data
        self._conn.sendall(data)
    
    def _flush(self):
        if self.buffered_transmit:
            self._conn.sendall(self.tx_buffer)
        self.tx_buffer = b""




from .serializer import Serializer

class BufferSerializer(Serializer):
    def __init__(self):
        self.tx_buffer = b""
        self.rx_buffer = b""

    def _receive(self, size):
        r = self.rx_buffer[:size]
        self.rx_buffer = self.rx_buffer[size:]
        return r
    
    def _transmit(self, data, size):
        self.tx_buffer += data

    def encode(self, data):
        self.tx_buffer = b""
        self.transmit(data)
        return self.tx_buffer
    
    def decode(self, data):
        self.rx_buffer = data
        return self.receive()

